/*==============================================================================
 $Id: alignment.c,v 1.1.1.1 2005/04/15 12:11:53 jkleinj Exp $ 
 alignment.c : Read FASTA sequences
 Copyright (C) 2004 Jens Kleinjung and John Romein
 see README file for Syntax, Copyright and License
==============================================================================*/

#include "aliss.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>


static int read_sequence_name(FILE *file, Seq *sequence)
{
    /* Reads a line starting with '>' and followed by the sequence name */
    /* Leading white space is skipped.  Reading proceeds until the end  */
    /* of the line is reached.  The name is read into sequence->name.   */

    int ch, name_length = 0, allocated = 64;

    while ((ch = getc(file)) != EOF && isspace(ch))
	;

    if (ch != '>')
	return 0;

    sequence->name = safe_malloc(allocated);

    do
	{
		sequence->name[name_length ++] = ch;

		if (name_length == allocated)
			sequence->name = safe_realloc(sequence->name, allocated += 64);
    } while ((ch = getc(file)) != EOF && ch != '\n' && isprint(ch));

    sequence->name[name_length] = '\0';
    return 1;
}


static void read_sequence_residues(FILE *file, Seq *sequence)
{
    /* Reads the residues in a sequence, up to (but not including) the      */
    /* next sequence header (starting with '>'), or up to end of file.      */
    /* Residues may span multiple lines.  White space and gaps are skipped. */
    /* Nonalpha characters are rejected, resulting in an error message.     */
    /* Alpha characters are converted to upper case.  The string is read    */
    /* into sequence->residues and zero-terminated; the length is stored    */
    /* into sequence->length.                                               */

    int ch, allocated = 64;

    sequence->res = safe_malloc(allocated);
    sequence->length = 0;

    while ((ch = getc(file)) != EOF && ch != '>')
	if (isalpha(ch))
	{
	    sequence->res[sequence->length ++] = toupper(ch);

	    if (sequence->length == allocated)
			sequence->res = safe_realloc(sequence->res, allocated += 64);
    }
	else if (ch != '.' && !isspace(ch)) 
	{
		fprintf(stderr, "Exiting: Illegal character '%c' in sequence\n", ch);
		exit(1);
    }

    if (ch == '>')
		ungetc('>', file);

    if (sequence->length == 0)
	{
		fprintf(stderr, "Exiting: Zero-sized sequence\n");
		exit(1);
    }

    sequence->res[sequence->length] = '\0';
}


static int read_sequence(FILE *file, Seq *sequence)
{
    if (read_sequence_name(file, sequence))
	{
		read_sequence_residues(file, sequence);
		return 1;
    } 
	else
		return 0;
}


void read_sequences(FILE *file, Ali *ali)
{ 
    ali->nseq	  = 0;
    ali->length   = 0;
    ali->sequence = safe_malloc(3 * sizeof(Seq));

    while (read_sequence(file, ali->sequence + ali->nseq))
	{
		if (ali->length < ali->sequence[ali->nseq].length)
			ali->length = ali->sequence[ali->nseq].length;

		++ ali->nseq;

		if (ali->nseq >= 3) /* allocate space for next sequence */
			ali->sequence = safe_realloc(ali->sequence, ali->nseq * sizeof(Seq));
    }

    if (ali->nseq < 2)
	{
		fprintf(stderr, "Exiting: Expected at least two sequences\n");
		exit(1);
    }

#ifdef VERBOSE
    fprintf(stdout, "Reference sequence length = %5d\n"
	                "Query     sequence length = %5d\n",
			ali->sequence[0].length, ali->sequence[1].length);
#endif
}
