/*=============================================================================
 $Id: aliss.c,v 1.6 2005/04/20 14:48:00 jkleinj Exp $

 ALISS: Alignment of Structures and Sequences
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	Tool for structure-structure, sequence-structure and sequence-sequence
	alignment.

 Copyright (C) 2003-2005 Jens Kleinjung
 Copyright (C) 2003-2005 Kuang Lin
 Copyright (C) 2004 John Romein
 Copyright (C) 2003 Arno Formella

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 see README file for Syntax, Copyright and License
==============================================================================*/

#include "aliss.h" 
#include "align.h" 
#include "alignment.h"
#include "cao.h"
#include "contactlist.h"
#include "contactmap.h"
#include "getsmat.h"
#include "kabsch.h"
#include "structure.h"
#include <gsl/gsl_vector_double.h>
#include <gsl/gsl_matrix_double.h>

float liftval	 = LIFTVAL; /* lift-value for CAO matrix */
float pamliftval = PAMLIFTVAL; /* lift-value for PAM matrix */
int   caoval     = CAOVAL; /* CAO distance */

/*----------------------------------------------------------------------------*/
/* file */
FILE *safe_open(const char *name, const char *mode)
{
    FILE *file = fopen(name, mode);

    if (file == 0)
	{
		fprintf(stderr, "Exiting: Cannot open file %s\n", name);
		exit(1);
    }

    return file;
}

/*----------------------------------------------------------------------------*/
/* allocation */
static void *check_non_null(void *ptr)
{
    if (ptr == 0)
	{
		fprintf(stderr, "Exiting: Cannot allocate memory\n");
		exit(1);
    }

    return ptr;
}

void *safe_malloc(size_t size)
{
    return check_non_null(malloc(size));
}

void *safe_realloc(void *ptr, size_t size)
{
    return check_non_null(realloc(ptr, size));
}

/*----------------------------------------------------------------------------*/
/* matrix */
float** allocate_matrix(int height, int width)
{
    int i;

    float **mat = safe_malloc(height * sizeof(float *));

    for (i = 0; i < height; ++ i)
		mat[i] = safe_malloc(width * sizeof(float));
		
    return mat;
}

void deallocate_matrix(float **mat, int height)
{
    int i;

    for (i = 0; i < height; i ++)
		free(mat[i]);

    free(mat);
}

void init_matrix(float **mat, int height, int width)
{
    int i, j;

    for (i = 0; i < height; ++ i)
		for (j = 0; j < width; ++ j)
			mat[i][j] = 0;
}

/*----------------------------------------------------------------------------*/
/* DP matrix */
float** init_dpmat(Ali *ali, float **smat)
{
	int i, j;
	float **dpmat;

    dpmat = allocate_matrix(ali->sequence[0].length, ali->sequence[1].length);
    init_matrix(dpmat, ali->sequence[0].length, ali->sequence[1].length);

	/* initialise DP matrix with substitution values */
    for (i = 0; i < ali->sequence[0].length; ++i)
		for (j = 0; j < ali->sequence[1].length; ++j)
			if ((ali->sequence[0].res[i] == '.') || (ali->sequence[1].res[j] == '.'))
				dpmat[i][j] = 0.;
			else
				dpmat[i][j] = smat[ali->sequence[0].res[i]-'A'][ali->sequence[1].res[j]-'A'];

	return dpmat;
}

void print_dpmat(float **dpmat, int lengthref, int lengthquery)
{
    int i, j;

    for (i=0; i<lengthref; ++i)
    {
		for (j=0; j<lengthquery; j++)
			fprintf(stderr, "%4.1f ", dpmat[i][j]);

		fprintf(stderr, "\n");
    }
}

/*----------------------------------------------------------------------------*/
/* print alignment */
void print_ali(FILE *outfile, Seq *seq0, Seq *seq1, char *p0, char *p1)
{
    fprintf(outfile, "%s\n%s\n%s\n%s\n", seq0->name, p0, seq1->name, p1);
}

/*----------------------------------------------------------------------------*/
/* position of substring in string */
int strpos(char *string, char *substring)
{
    char *a, *b;
    int pos = 0;

    /* first find single character match, then match rest of substring */

    b = substring;
    if (*b == 0)
        return 0;

    for ( ; *string != 0; string += 1)
    {
        ++pos;
        if (*string != *b)
            continue;

        a = string;
        while (1)
        {
            if (*b == 0)
                return pos;

            if (*a++ != *b++)
                break;
        }
        b = substring;
    }
    return pos;
}

/*----------------------------------------------------------------------------*/
/* get termini */
void get_termini(Ali *ali, char *p0, char *p1)
{
    int i, j;
    char nogap0[strlen(p0)+1];
    char nogap1[strlen(p1)+1];

   /* no-gap aligned sequence 0 */
    for (i = 0, j = 0; j < strlen(p0); ++j)
        if (p0[j] != '.')
			nogap0[i++] = p0[j];
	nogap0[i] = '\0';
	
    /* no-gap aligned sequence 1 */
    for (i = 0, j = 0; j < strlen(p1); ++j)
        if (p1[j] != '.')
            nogap1[i++] = p1[j];
	nogap1[i] = '\0';

	/* safety loop */
	if (strlen(nogap0) == 0 || strlen(nogap1) == 0)
	{
		fprintf(stderr, "Exiting: Zero length sequence string\n");
		exit(1);
	}

    /* termini of aligned sequence in total sequence */
    ali->sequence[0].aliN = strpos(ali->sequence[0].res, nogap0) - 1;
    ali->sequence[0].aliC = ali->sequence[0].aliN + strlen(nogap0) - 1;
    ali->sequence[1].aliN = strpos(ali->sequence[1].res, nogap1) - 1;
    ali->sequence[1].aliC = ali->sequence[1].aliN + strlen(nogap1) - 1;

	/* safety loop */
	if (ali->sequence[0].aliN < 0 || ali->sequence[1].aliN < 0)
	{
		fprintf(stderr, "Exiting: N-terminus invalid\n");
		exit(1);
	}

#ifdef VERBOSE
    fprintf(stdout, "Termini mapped onto input sequences: seq0 = [%d-%d], seq1 = [%d-%d]\n",
        ali->sequence[0].aliN, ali->sequence[0].aliC, 
		ali->sequence[1].aliN, ali->sequence[1].aliC);
#endif
}

/*----------------------------------------------------------------------------*/
/* get aligned positions (no gaps) */
/* these positions will be used for structural superpositioning */
void get_aligned_pos(Ali *ali, char *p0, char *p1)
{
	int i;
	int match = 0; /* match = gapless alignment position */
	int npos0 = 0; /* residue count */
	int npos1 = 0;

	ali->sequence[0].pos = safe_malloc((ali->length + 1) * sizeof(int));
	ali->sequence[1].pos = safe_malloc((ali->length + 1) * sizeof(int));

	/* store aligned positions */
	for (i = 0; i < strlen(p0); ++i)
	{
		if (p0[i] != '.' && p1[i] != '.')
		{
			++match;
			/* store matched residue numbers (referring to original sequence) */
			ali->sequence[0].pos[match] = npos0 + ali->sequence[0].aliN;
			ali->sequence[1].pos[match] = npos1 + ali->sequence[1].aliN;
			++npos0;
			++npos1;
		}
		if (p0[i] != '.' && p1[i] == '.')
			++npos0;
		if (p0[i] == '.' && p1[i] != '.')
			++npos1;
	}

	/* store number of aligned positions at array position 0 */
	ali->sequence[0].pos[0] = ali->sequence[1].pos[0] = match;
}

/*----------------------------------------------------------------------------*/
/* sequence identity calculation of aligned gap-free positions */
float identity(char *p0, char *p1)
{
	int i = 0;
	int gap = 0;
	float id = 0; /* number of identical positions */

	while ((p0[i] != '\0') && (p1[i] != '\0'))
	{
		if ((p0[i] == '.') || (p1[i] == '.'))
			++gap;
		if (p0[i] == p1[i])
			++id;
		++i;
	}
	return id /= (i - gap);
}

/*----------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{ 
    FILE *alignfile; char *alignfilename = 0;
    FILE *contactfile; char *contactfilename = 0;
    FILE *pdbfile0; char *pdbfilename0 = 0;
    FILE *pdbfile1; char *pdbfilename1 = 0;
    FILE *outfile; char *outfilename = OUTFILENAME;
    FILE *pdboutfile; char *pdboutfilename = PDBOUTFILENAME;

    int i = 0;
	int j = 0;
	int k = 0;

	/* flags for command line parameters */
    int setl = 0, setm = 0, seto = 0, setu = 0, setw = 0, setp = 0, setq = 0;
    float inc_val;

    extern char *optarg; /* variables used by getopt */
    int clp;

    Ali ali; /* alignment information */
	Contl cont; /* contact list information */
	Str pdb0; /* structure 0 */
	Str pdb1; /* structure 1 */

	float **dpmat = 0; /* DP matrix */
	float ****caomat = 0; /* CAO matrix */
	gsl_matrix *U = 0; /* rotation matrix */
	gsl_vector *t = 0; /* translation vector */

	int reforque; /* flag for reference or query sequence */

	int cref0, cref1, cque0, cque1; /* position of contact residues */
	int rref0, rref1, rque0, rque1; /* contact residues */

	int con; /* this contact */
	int que; /* this reference and query position */

	int global = GLOBAL; /* local/global alignment switch */
	float wcao = WCAO; /* weight of CAO scores versus PAM scores */
	float gapo = GAPO, gape = GAPE; /* gap opening and gap extension penalties */
	int allatom = ALLATOM; /* all atom output switch */

	char *str0, *str1; /* space to store alignment */

	double *s = 0; /* the optimal scaling, if set != 0 */
	float rmsd = 0; /* rmsd value of superimposed structures */
	float stsc = 0; /* structure score */
	float stscf = 0.08; /* BENCHMARK, structure score factor */
	int pep = 5; /* length of peptide to superimpose */
	int shift = pep; /* peptide shift */
	int minlength = 0; /* length of shorter protein structure */


#ifdef VERBOSE
	fprintf(stdout, 
	"\t================================================\n"
	"\t  ALISS: ALIgnment of Structures and Sequences\n"
	"\t      Release $Date: 2005/04/20 14:48:00 $\n"
	"\t================================================\n");
#endif

	/*----------------------------------------------------------------------------*/
	/* get command line arguments */
	while((clp = getopt(argc,argv, "a:c:o:u:l:m:v:p:q:w:x:y:zgh")) != -1)
	{
		switch(clp)
		{
			/* input alignment */
			case 'a':
				alignfilename = optarg;
#ifdef VERBOSE
				fprintf(stdout, "Input alignment: %s\n", alignfilename);
#endif
			break;

			/* input first contact list */
			case 'c':
				contactfilename = optarg;
#ifdef VERBOSE
				fprintf(stdout, "Input contactlist: %s\n", contactfilename);
#endif
			break;

			/* use global alignment */
			case 'g':
				global = 1;
#ifdef VERBOSE
				fprintf(stdout, "Global alignment\n");
#endif
			break;

			/* CAO matrix lift-value */
			case 'l':
				liftval = atof(optarg);
#ifdef VERBOSE
				fprintf(stdout, "Matrix lift-value: %6.2f\n", liftval);
#endif
				++setl;
			break;

			/* PAM matrix lift-value */
			case 'm':
				pamliftval = atof(optarg);
#ifdef VERBOSE
				fprintf(stdout, "Matrix lift-value: %6.2f\n", pamliftval);
#endif
				++setm;
			break;

			/* alignment output filename */
			case 'o':
				outfilename = optarg;
#ifdef VERBOSE
				fprintf(stdout, "Sequence alignment output file: %s\n", outfilename);
#endif
				++seto;
			break;

			/* gap opening penalty */
			case 'p':
				gapo = atof(optarg);
#ifdef VERBOSE
				fprintf(stdout, "Gap opening penalty: %4.2f\n", gapo);
#endif
				++setp;
			break;

			/* gap extension penalty */
			case 'q':
				gape = atof(optarg);
#ifdef VERBOSE
				fprintf(stdout, "Gap extension penalty: %4.2f\n", gape);
#endif
				++setq;
			break;

			/* PDB output filename */
			case 'u':
				pdboutfilename = optarg;
#ifdef VERBOSE
				fprintf(stdout, "Structure alignment output file: %s\n", pdboutfilename);
#endif
				++setu;
			break;
	    
			/* CAO value */
			case 'v':
				caoval = atoi(optarg);
#ifdef VERBOSE
				fprintf(stdout, "CAO value: %d\n", caoval);
#endif
			break;
	    
			/* weight CAO/PAM */
			case 'w':
				wcao = atof(optarg);
#ifdef VERBOSE
				fprintf(stdout, "Weight CAO scores versus PAM scores: %4.2f\n", wcao);
#endif
				++setw;
			break;

			/* PDB file 0 */
			case 'x':
				pdbfilename0 = optarg;
#ifdef VERBOSE
				fprintf(stdout, "PDB file 0: %s\n", pdbfilename0);
#endif
			break;

			/* PDB file 1 */
			case 'y':
				pdbfilename1 = optarg;
#ifdef VERBOSE
				fprintf(stdout, "PDB file 1: %s\n", pdbfilename1);
#endif
			break;
			/* all atom */
			case 'z':
				allatom = 1;
#ifdef VERBOSE
				fprintf(stdout, "All atom output\n");
#endif
			break;

			/* help */
			case 'h':
				fprintf(stdout, "\n\tSyntax:\n"
						"\t\taliss -x <structure file0 [PDB]> -y <structure file1 [PDB]>\n"
						"\t\taliss -a <seqs-to-align file [FASTA]> -c <contact-list file0 [FORMATTED]>\n"
						"\t\taliss -a <seqs-to-align file [FASTA]>\n"
);
				fprintf(stdout, "\n\tOptions:\n"
						"\t\t-a <seqs-to-align file [FASTA]> input alignment\n"
						"\t\t-c <contact-list file [FORMATTED]> input contacts first sequence\n"
						"\t\t-g global alignment (default local)\n"
						"\t\t-h Help\n"
						"\t\t-l <lift-value [float]> add constant to CAO matrix values (default %4.2f)\n"
						"\t\t-m <lift-value [float]> add constant to PAM matrix values (default %4.2f)\n"
						"\t\t-o <output file> sequence alignment (default '%s')\n"
						"\t\t-p <gap opening penalty [float]>(default %4.2f)\n"
						"\t\t-q <gap extension penalty [float]> (default %4.2f)\n"
						"\t\t-u <output file> structure alignment (default %s)\n"
						"\t\t-v <CAO distance> (default %d)\n"
						"\t\t-w <weight [float]> weight of CAO scores versus PAM scores (default %4.2f)\n"
						"\t\t-x <structure file [PDB]> input structure 0\n"
						"\t\t-y <structure file [PDB]> input structure 1\n"
						"\t\t-z all atom structure alignment invoking side chain contacts (default Calpha only)\n",
							LIFTVAL, PAMLIFTVAL, OUTFILENAME, GAPO, GAPE,
							PDBOUTFILENAME, CAOVAL, WCAO);
				fprintf(stdout, "\n\tSee 'README' file for Description, Copyright and License\n\n");
			exit(1);
		}
	}

    /*----------------------------------------------------------------------------*/
	/* some checks */
    if (!alignfilename && (!pdbfilename0 || !pdbfilename1))
	{
		fprintf(stderr, "Exiting: Need at least one alignment file or two structure files!\n");
		fprintf(stderr, "'aliss -h' for Help\n");
		exit(1);
    }

	if (!pdbfilename0 && pdbfilename1)
	{
		fprintf(stderr, "Exiting: Need second structure for structure-structure alignment.\n");
		fprintf(stderr, "'aliss -h' for Help\n");
		exit(1);
	}

	if (contactfilename && pdbfilename0 && pdbfilename1)
		fprintf(stderr, "Warning: Specified contact list will be ignored\n");


    /*----------------------------------------------------------------------------*/
	/* print input information */
#ifdef VERBOSE
    if (seto == 0)
		fprintf(stdout, "Default sequence alignment output file: %s\n", outfilename);

	if (setu == 0 && pdbfilename0 && pdbfilename1)
		fprintf(stdout, "Default structure alignment output file: %s\n", pdboutfilename);

    if (global == 0)
		fprintf(stdout, "Default local alignment\n");

    if (setp == 0)
		fprintf(stdout, "Default gap opening penalty: %4.2f\n", gapo);

    if (setq == 0)
		fprintf(stdout, "Default gap extension penalty: %4.2f\n", gape);

    if (setl == 0 && contactfilename)
		fprintf(stdout, "Default CAO matrix constant: %4.2f\n", liftval);

    if (setw == 0 && contactfilename)
		fprintf(stdout, "Default CAO score weight: %4.2f\n", wcao);

    if (caoval == CAOVAL && contactfilename)
		fprintf(stdout, "Default CAO value: %d\n", caoval);
#endif

    /*----------------------------------------------------------------------------*/
    /* PAM matrix */
    float **smat = allocate_matrix(32, 32); /* make PAM matrix */
    get_smat(smat, "Blosum62", pamliftval); /* load matrix */

    /*----------------------------------------------------------------------------*/
#ifdef VERBOSE
	if (!pdbfilename0 && !pdbfilename1 && !contactfilename)
		fprintf(stdout, "\nSEQUENCE-SEQUENCE\n");
	if (contactfilename && !pdbfilename1)
		fprintf(stdout, "\nSEQUENCE-STRUCTURE\n");
	if (pdbfilename0 && pdbfilename1)
		fprintf(stdout, "\nSTRUCTURE-STRUCTURE\n");
#endif


    /*----------------------------------------------------------------------------*/
	/* STRUCTURE-STRUCTURE */
	/* perform structural scoring if PDB files are given */
	if (pdbfilename0 && pdbfilename1)
	{
		/* read PDB files */
		ali.sequence = safe_malloc(2 * sizeof(Seq));
		/* template */
		pdbfile0 = safe_open(pdbfilename0, "r");
		pdb0.sequence.name = safe_malloc((strlen(pdbfilename0) + 1) * sizeof(char));
		strcpy(pdb0.sequence.name, pdbfilename0);
		ali.sequence[0].name = safe_malloc((strlen(pdbfilename0) + 1) * sizeof(char));
		strcpy(ali.sequence[0].name, pdbfilename0);
		read_pdb(pdbfile0, &pdb0);
		if (allatom)
			read_all_pdb(pdbfile0, &pdb0);
		fclose(pdbfile0);

		/* query */
		pdbfile1 = safe_open(pdbfilename1, "r");
		pdb1.sequence.name = safe_malloc((strlen(pdbfilename1) + 1) * sizeof(char));
		strcpy(pdb1.sequence.name, pdbfilename1);
		ali.sequence[1].name = safe_malloc((strlen(pdbfilename1) + 1) * sizeof(char));
		strcpy(ali.sequence[1].name, pdbfilename1);
		read_pdb(pdbfile1, &pdb1);
		if (allatom)
			read_all_pdb(pdbfile1, &pdb1);
		fclose(pdbfile1);

		/* assign sequences of specified structures */
		assign_seq(&ali, &pdb0, &pdb1);
		ali.length = ali.sequence[0].length > ali.sequence[1].length ? ali.sequence[0].length : ali.sequence[1].length;

		/* DP matrix : initialise and fill with 'smat' Dayhoff-type substitution values */
		dpmat = init_dpmat(&ali, smat);

		/* allocate memory to symmetry operators for structural superpositioning */
		U = gsl_matrix_alloc(3, 3); /* rotation matrix */
		t = gsl_vector_alloc(3); /* translation vector */

		/* determine size range of peptides */
		/* 'peptide' means fragment of query structure */
		minlength = pdb0.natom < pdb1.natom ? pdb0.natom : pdb1.natom;

#ifdef VERBOSE
		fprintf(stdout, "Superpositioning: Scanning with fragment size %d and shift %d\n", 
				pep, shift);
#endif
		/* allocate memory to atom matrices */
		gsl_matrix *X = gsl_matrix_alloc(pep, 3); /* query coordinates */
		gsl_matrix *Y = gsl_matrix_alloc(pep, 3); /* template coordinates */

		/* 3D alignment: scan for good matches using peptides */
		for (j = 0; j <= pdb1.natom - pep - shift; j += shift) /* scan with peptides */ 
			for (i = 0; i <= pdb0.natom - pep; ++i) /* each position of template */
			{
				def_points(i, j, pep, X, Y, &pdb0, &pdb1); /* define points to superimpose */		
				kabsch(pep, X, Y, U, t, s); /* Kabsch superpositioning routine */
				rmsd = transform_coordinates(i, j, pep, U, t, &pdb0, &pdb1); /* transform x' = Ux + t */
				stsc = struc_score(pep, rmsd, stscf); /* calculate structure score */
				/*fprintf(stdout,"%f ", stsc);*/

				/* add structure score to DP matrix normalised by repeated scoring */
				for (k = 0; k < pep; ++k)
					dpmat[i+k][j+k] += stsc;
			}
		gsl_matrix_free(X);
		gsl_matrix_free(Y);
	}

    /*----------------------------------------------------------------------------*/
    /* SEQUENCE-SEQUENCE or SEQUENCE-STRUCTURE */
    /* read input sequences if given */
	if (alignfilename)
	{
		alignfile = safe_open(alignfilename, "r");
		read_sequences(alignfile, &ali);
		fclose(alignfile);
	}

    /*----------------------------------------------------------------------------*/
	/* SEQUENCE-STRUCTURE */
	/* map contacts from given structure onto sequence with unknown structure */
    /* for all matching contacts, add CAO scores to DP matrix */

	/* DP matrix : initialise and fill with 'smat' Dayhoff-type substitution values */
	if (! dpmat)
		dpmat = init_dpmat(&ali, smat);

    /* read contact list if given */
	if (contactfilename)
	{
		contactfile = safe_open(contactfilename, "r");
		reforque = 0;
		contactlist(contactfile, &cont, &ali, reforque);
		fclose(contactfile); 
	}

	/* perform contact-based scoring */
	if (contactfilename && !pdbfilename1)
	{
		/* CAO matrix */
		caomat = make_caomat(); /* make CAO matrix */
		cao_matrix(caomat, caoval, liftval); /* load CAO matrix */

		/* for all contacts, add CAO scores to DP matrix */
		for (con = 0; con < cont.ncontact; ++ con)
		{
			/* scan over all query positions */
			for (que = 0; que < ali.sequence[1].length; ++ que)
			{
				cref0 = cont.contact[con][0]; /* reference positions */
				cref1 = cont.contact[con][1];

				rref0 = ali.sequence[0].res[cref0]; /* reference residues */
				rref1 = ali.sequence[0].res[cref1];

				cque0 = que; /* query positions */
				cque1 = cque0 + (cont.contact[con][1] - cont.contact[con][0]);

				if ((cque1 < 0) || (cque1 >= ali.sequence[1].length))
					continue;

				rque0 = ali.sequence[1].res[cque0]; /* query residues */
				rque1 = ali.sequence[1].res[cque1];

				inc_val = wcao * caomat[rref0-'A'][rref1-'A'][rque0-'A'][rque1-'A'];
				dpmat[cref0][cque0] += inc_val;
				dpmat[cref1][cque1] += inc_val;
			}
		}
    } 

    /* print DP matrix */
    /*print_dpmat(dpmat, ali.sequence[0].length, ali.sequence[1].length);*/

    /*----------------------------------------------------------------------------*/
	/* DP ALIGNMENT */
    /* perform sequence alignment */
    str0 = safe_malloc(2 * ali.length);
    str1 = safe_malloc(2 * ali.length);
    align(dpmat, &ali.sequence[0], &ali.sequence[1], str0, str1, global, gapo, gape);
    
    /* print sequence alignment */
    outfile = safe_open(outfilename, "w");
    print_ali(outfile, &ali.sequence[0], &ali.sequence[1], str0, str1);
    fclose(outfile);
    get_termini(&ali, str0, str1);
#ifdef VERBOSE
	fprintf(stdout, "Alignment sequence identity = %f\n", identity(str0, str1));
#endif

    /*----------------------------------------------------------------------------*/
    /* final structure superpositioning */
	if (pdbfilename0 && pdbfilename1)
	{
		/* perform 3D alignment on the basis of the matched residues */
		/* determine number of atoms to superimpose = aligned positions */
		get_aligned_pos(&ali, str0, str1);
		pep = ali.sequence[1].pos[0];

		/* perform superpositioning */	
		/* allocate memory to atom matrices */
		gsl_matrix *X = gsl_matrix_alloc(pep, 3); /* query coordinates */
		gsl_matrix *Y = gsl_matrix_alloc(pep, 3); /* template coordinates */

		/* 3D alignment of aligned positions */
		/* define points to superimpose */
		def_ali_points(&ali, pep, X, Y, &pdb0, &pdb1); 		
		kabsch(pep, X, Y, U, t, s); /* Kabsch superpositioning routine */
		/* transform x' = Ux + t */
		rmsd = transform_ali_coordinates(&ali, U, t, &pdb0, &pdb1, allatom);
		stsc = struc_score(pep, rmsd, stscf); /* calculate structure score */
#ifdef VERBOSE
		fprintf(stdout, "Structure alignment score = %f for length = %d and rmsd = %f\n", 
				stsc, pep, rmsd);
#endif
#ifdef BENCHMARK
		fprintf(stdout, "%8.4f  %8.4f  %8.4f  %3d %3d\t%25s\t", 
			identity(str0, str1), stsc, rmsd, pep, ali.length, pdbfilename0);
#endif
		/* print 3D alignment */
		pdboutfile = safe_open(pdboutfilename, "w");
		write_pdb(pdboutfile, &pdb0, &pdb1, &ali, allatom);
		fclose(pdboutfile);

		/* free matrices */
		gsl_matrix_free(X);
		gsl_matrix_free(Y);
		gsl_matrix_free(U);
		gsl_vector_free(t);
	}

	fprintf(stdout, "\n");

    /*----------------------------------------------------------------------------*/
	/* free memory */

	/* free matrices */
    deallocate_matrix(dpmat, ali.sequence[0].length);
    deallocate_matrix(smat, 32);
	if (contactfilename) delete_caomat(caomat);

	/* free structure */
	if (pdbfilename0)
	{
		free(pdb0.atom);
		if (allatom) free(pdb0.allatom);
		free(pdb0.sequence.name);
		free(pdb0.sequence.res);
		free(ali.sequence[0].pos);
	}
    if (pdbfilename1)
	{
		free(pdb1.atom);
		if (allatom) free(pdb1.allatom);
		free(pdb1.sequence.name);
		free(pdb1.sequence.res);
		free(ali.sequence[1].pos);
	}

	/* free contact */
	if (contactfilename)
	{
		free(cont.contact);
		free(cont.ncontres);
	}

	/* free sequence */
    free(str0), free(str1);
    free(ali.sequence[0].res);
    free(ali.sequence[1].res);
    free(ali.sequence[0].name);
    free(ali.sequence[1].name);

    free(ali.sequence);

	return 0;
}

