/*===============================================================================
 $Id: aliss.h,v 1.1.1.1 2005/04/15 12:11:53 jkleinj Exp $
 ALISS: Alignment of Structures and Sequences
 aliss.h : Main routine
 Copyright (C) 2004-2005  Jens Kleinjung, Amsterdam London
 see README file for Syntax, Copyright and License
================================================================================*/

#if !defined ALISS_H
#define ALISS_H

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>

extern float liftval;

#define VERBOSE
/*#define BENCHMARK*/

/*-----------------------------------------------------------------------------*/
/* defaults */

#define CAOVAL 120
#define OUTFILENAME "aliss.ali" /* default output filename */
#define PDBOUTFILENAME "aliss.pdb" /* default output filename */
#define GLOBAL 0 /* local/global alignment switch */
#define LIFTVAL 2.8 /* default CAO matrix lift value */
#define PAMLIFTVAL 0.0 /* default PAM matrix lift value */
#define WCAO 0.1 /* weight of CAO scores versus PAM scores */
#define GAPO 14.0 /* gap opening penalty in DP */
#define GAPE 1.0 /* gap extension penalty in DP */
#define ALLATOM 0 /* all atom output */
#define MAXCONTDIST 8 /* maixmal contact distance */

/* CAO contact matrix */
#define NACID 26 /* number of amino acids (really alphabet letters) */
#define NCAO 20 /* number of amino acid residues specified by CAO matrix */
#define NNOT 6 /* number of not specified characters */

/*-----------------------------------------------------------------------------*/
/* structures */

/* vector is array of floats */
typedef struct
{
   float x, y, z;
} Vec;

/* sequence */
typedef struct  
{
    char *name; /* sequence name */
    char *res; /* array of residues = sequence */
    int length; /* length of sequence */
	int aliN; /* N-terminal position in alignment */
	int aliC; /* C-terminal position alignment */
	int *pos; /* aligned positions */
} Seq;

/* sequence alignment */
typedef struct
{
    Seq *sequence; /* array of sequences constituting sequence alignment */
    int nseq; /* number of sequences in alignment */
    int length; /* length of sequence alignment */
} Ali;

/* contact list */
typedef struct
{
    int (*contact)[2]; /* contact list for ungapped sequences */
    int ncontact; /* number of contacts */
    int *ncontres; /* number of contacts per residue */
    int maxcn; /* maximal number of contacts per residue */
    float avgcn; /* average number of contacts per residue */
} Contl;

/* contact map */
typedef struct
{
    int (*contact)[2]; /* Cb contact list for structure */
    int ncontact; /* number of contacts */
	int (*cblist)[2]; /* list of Cb atoms */
    Vec *midpoint; /* Cb contact midpoints */
	int (*contlim)[2]; /* first/last contact of residue */
    int (*contmatch)[2]; /* contact match list for template/query  */
	int nmatch; /* number of contact matches */
} Contm;

/* atom : definition of PDB atom format, numbers indicate columns */
/* only the entries used in this program are activated so save resources */
typedef struct
{
	char record_ne[8]; /* Record type; 1 -  6*/
	int atom_nr; /* Atom serial number;  7 - 11 */
	char atom_ne[8]; /* Atom name; 13 - 16 */
	/*char alt_loc[2];*/ /* Alternate location indicator; 17 */
	char res_ne[4]; /* Residue name; 18 - 20 */
	char chain_id[2]; /* Chain identifier; 22 */
	/*int res_nr;*/ /* Residue sequence number; 23 - 26 */
	/*char icode[2];*/ /* Code for insertion of residues; 27 */
	float x; /* Orthogonal coordinates for X in Angstroms; 31 - 38 */
	float y; /* Orthogonal coordinates for Y in Angstroms; 39 - 46 */
	float z; /* Orthogonal coordinates for Z in Angstroms; 47 - 54 */
	/*float occupancy;*/ /* Occupancy; 55 - 60 */
	/*float temp_f;*/ /* Temperature factor; 61 - 66 */
	/*char seg_id[5];*/ /* Segment identifier; 73 - 76 */
	/*char element[3];*/ /* Element symbol; 77 - 78 */
	/*char charge[3];*/ /* Charge on the atom; 79 - 80 */
	char descrip[32]; /* everything before coordinates */
	float tx; /* transformed x : tx = U * x + t */
	float ty; /* transformed y : ty = U * y + t */
	float tz; /* transformed z : tz = U * z + t */
} Atom;

/* structure is array of atoms */
typedef struct
{
	Atom *atom; /* array of selected (CA) atoms constituting structure */
	int natom; /* number of selected (CA) atoms */
	Atom *allatom; /* array of all atoms constituting structure */
	int nallatom; /* number of all selected (CA) atoms */
	Seq sequence; /* sequence of structure */
	Contm contmap; /* contact map */
} Str;

/* structure alignment */
typedef struct
{
    Str *structure; /* array of structures constituting alignment */
    int nstr; /* number of structures in alignment */
    int length; /* length of structure alignment */
} Strali;


extern float vector_len_sq(Vec *);
extern float vector_len(Vec *);
extern Vec *vector_add(Vec *, Vec *, Vec *);
extern Vec *vector_subtract(Vec *, Vec *, Vec *);
extern Vec *vector_scale(Vec *, float);

extern void *safe_malloc(size_t), *safe_realloc(void *, size_t);
extern float **allocate_matrix(int height, int width);
extern void deallocate_matrix(float **, int height);
extern void init_matrix(float **mat, int height, int width);
float** init_dpmat(Ali *ali, float **smat);
void get_aligned_pos(Ali *ali, char *p0, char *p1);
FILE *safe_open(const char *name, const char *mode);
void print_dpmat(float **dpmat, int lengthref, int lengthquery);

#endif
