/*==============================================================================
 $Id: cao.c,v 1.1.1.1 2005/04/15 12:11:53 jkleinj Exp $ 
 aliss.c : CAO matrix conversions
 Copyright (C) 2004 John Romein
 see README file for Syntax, Copyright and License
===============================================================================*/

#include "aliss.h"
#include "cao.h"
#include "ncao.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


/* Allocate CAO 4-D matrix.  The 4-D matrix is stored as a tree (or rather,
   a directed graph).  Save 85% on memory usage by mapping all entries
   that contain at least one non-residue value to the same array
   (non_residues[]).  Allocation is done recursively. */

static const char CAOR[] = "ARNDCQEGHILKMFPSTWYV";
static float	  non_residues[NACID];
static void	  *indirect[2][NACID];
static void	  *nonres[3] = { non_residues, indirect, indirect + 1 };


/*---------------------------------------------------------------------------*/
static void *allocate_indirect(int depth)
{
    int i;

    if (depth == 0) {
	float *array = safe_malloc(NACID * sizeof(float));

	for (i = 0; i < NACID; i ++)
	    array[i] = liftval;

	for (i = 0; CAOR[i] != '\0'; i ++)
	    array[CAOR[i] - 'A'] = 0;

	return array;
    } else {
	void **array = safe_malloc(NACID * sizeof(void *));

	for (i = 0; i < NACID; i ++)
	    array[i] = nonres[depth - 1];

	for (i = 0; CAOR[i] != '\0'; i ++)
	    array[CAOR[i] - 'A'] = allocate_indirect(depth - 1);

	return array;
    }
}

/*---------------------------------------------------------------------------*/
float ****make_caomat(void)
{
    int i, depth;

    for (i = 0; i < NACID; i ++)
	non_residues[i] = liftval;

    for (depth = 0; depth < 2; depth ++)
	for (i = 0; i < NACID; i ++)
	    indirect[depth][i] = nonres[depth];

    return allocate_indirect(3);
}

/*---------------------------------------------------------------------------*/
static void deallocate_indirect(void *array, int depth)
{
    int i;

    if (depth != 0)
	for (i = 0; CAOR[i] != '\0'; i ++)
	    deallocate_indirect(((void **) array)[CAOR[i] - 'A'], depth - 1);

    free(array);
}

/*---------------------------------------------------------------------------*/
void delete_caomat(float ****caomat)
{
    deallocate_indirect(caomat, 3);
}

static vector L, LV;
static matrix CAO;

/*---------------------------------------------------------------------------*/
static void matrix_multiply(matrix d, matrix s1, matrix s2)
{
    int i, j, k;

    for (i = 0; i < MATRIX_SIZE; i ++)
	for (j = 0; j < MATRIX_SIZE; j ++) {
	    float sum = 0;

	    for (k = 0; k < MATRIX_SIZE; k ++)
		sum += s1[i][k] * s2[k][j];

	    d[i][j] = sum;
	}
}

/*---------------------------------------------------------------------------*/
static void compute_cao(int caoval)
{
    int    i, j;
    float  t;

    for (i = 0; i < MATRIX_SIZE; i ++) {
	L[i]  = exp(caoval * d[i]);
	/*LV[i] = log_res_ocu[i % 20] + log_res_ocu[i / 20];*/ /* str-seq */
	LV[i] = log_cont_ocu[i]; /* str-str */
    }

    for (i = 0; i < MATRIX_SIZE; i ++)
	for (j = 0; j < MATRIX_SIZE; j ++)
	    S[i][j] *= L[j];

    matrix_multiply(CAO, S, S1);

    for (i = 0; i < MATRIX_SIZE; i ++) {
	t = 0;

	for (j = 0; j < MATRIX_SIZE; j ++)
	    t += CAO[i][j];

	t = cont_ocu[i] / t;

	for (j = 0; j < MATRIX_SIZE; j ++)
	    CAO[i][j] *= t;
    }

    for (i = 0; i < MATRIX_SIZE; i ++)
	for (j = i + 1; j < MATRIX_SIZE; j ++)
	    CAO[i][j] = CAO[j][i] = (CAO[i][j] + CAO[j][i]) / 2;

    for (i = 0; i < MATRIX_SIZE; i ++) {
	t = 1 / cont_ocu[i];

	for (j = 0; j < MATRIX_SIZE; j ++)
	    CAO[i][j] = log(CAO[i][j] * t) - LV[j];
    }
}

/*---------------------------------------------------------------------------*/
static void get_cao(int caoval)
{
    char cache_file_name[32];
    FILE *cache_file;

    sprintf(cache_file_name, "/tmp/cao%dcache", caoval);

    if ((cache_file = fopen(cache_file_name, "rb")) != 0 && fread(CAO, sizeof CAO, 1, cache_file) != 1)
	{
		fprintf(stderr, "Warning: CAO cache file corrupted\n");
		fclose(cache_file);
		cache_file = 0;
		remove(cache_file_name);
    }

    if (cache_file == 0)
	{
		compute_cao(caoval);

		if ((cache_file = fopen(cache_file_name, "wb")) == 0 || fwrite(CAO, sizeof CAO, 1, cache_file) != 1)
			fprintf(stderr, "Warning: could not write CAO cache file\n");
    }

    if (cache_file != 0)
	fclose(cache_file);
}

/*---------------------------------------------------------------------------*/
void cao_matrix(float ****caomat, int caoval, float liftval)
{
    int a, b, c, d, i, j, k, l;

    /* define matrix AA order of CAO */
    const char CAOR[21] = "ARNDCQEGHILKMFPSTWYV";
    const char CNOT[7]  = "BJOUXZ";

  /* assign matrix values */

  /* this routine assigns the matrix values above to the matrix 'caomat' */
  /* 1. dimension: res 1 of contact 1 */
    get_cao(caoval);

  for (i=0; i<NCAO; i++)
    {
      /* convert counter 'i' to respective residue value in CAOR (in integer range [0-25]) */
      a = CAOR[i]-'A';
      /* 2. dimension: res 2 of contact 1 */
      for (j=0; j<NCAO; j++)
        {
          b = CAOR[j]-'A';
          /* 3. dimension: res 1 of contact 2 */
          for (k=0; k<NCAO; k++)
            {
              c = CAOR[k]-'A';
              /* 4. dimension: res 2 of contact 2 */
              for (l=0; l<NCAO; l++)
                {
                  d = CAOR[l]-'A';
                  /* the matrix is caomat[i][j][k][l], but the correct assignment to the */
                  /* ... matrix-specific residue order is achieved by caomat[a][b][c][d] */
                  /* printf("[%d %d %d %d] = %4.2f\n", a,b,c,d, caomat[a][b][c][d]); */
                  caomat[a][b][c][d] = CAO[NCAO * i + j][NCAO * k + l] + liftval;
                }
            }
        }
    }

  /* this routine assigns the matrix values of characters not specified by CAO */
  /* in case these appear in a sequence their value will be zero */
  /* 1. dimension: res 1 of contact 1 */
  for (i=0; i<NNOT; i++)
    {
      /* convert counter 'i' to respective residue value in CAOR (in integer range [0-25]) */
      a = CNOT[i]-'A';
      /* 2. dimension: res 2 of contact 1 */
      for (j=0; j<NNOT; j++)
        {
          b = CNOT[j]-'A';
          /* 3. dimension: res 1 of contact 2 */
          for (k=0; k<NNOT; k++)
            {
              c = CNOT[k]-'A';
              /* 4. dimension: res 2 of contact 2 */
              for (l=0; l<NNOT; l++)
                {
                  d = CNOT[l]-'A';
                  /* assign value of zero to non-specified characters */
                  caomat[a][b][c][d] = 0;
                  /* printf("[%d %d %d %d] = %4.2f\n", a,b,c,d, caomat[a][b][c][d]); */
                }
            }
        }
    }
}
