/*==============================================================================
 $Id: contactlist.c,v 1.1.1.1 2005/04/15 12:11:53 jkleinj Exp $
 contactlist.c : Read precompiled list of contacts (derived from PDB file)
 Copyright (C) 2004 Jens Kleinjung and John Romein
 see README file for Syntax, Copyright and License
===============================================================================*/

#include "aliss.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>


static void read_reference_sequence(FILE *contactfile, Ali *ali)
{
    /* first line lists contact sequence */

    int ch, i = 0;

    while (i < ali->sequence[0].length)
	if ((ch = getc(contactfile)) == EOF || (!isspace(ch) && toupper(ch) != ali->sequence[0].res[i ++])) {
	    fprintf(stderr, "Warning: First sequence and contact list 1 disagree at position %d\n", i);
	    exit(1);
	}
}

static void read_query_sequence(FILE *contactfile, Ali *ali)
{
    /* first line lists contact sequence */

    int ch, i = 0;

    while (i < ali->sequence[1].length)
	if ((ch = getc(contactfile)) == EOF || (!isspace(ch) && toupper(ch) != ali->sequence[1].res[i ++])) {
	    fprintf(stderr, "Exiting: Second sequence and contact list 2 disagree at position %d\n", i);
	    exit(1);
	}
}

static void read_nr_contacts(FILE *contactfile, Contl *cont)
{
    /* second line contains number of contacts */

    if (fscanf(contactfile, "%d\n", &cont->ncontact) != 1 || cont->ncontact <= 0) {
	fprintf(stderr, "Exiting: expected number of contacts on second line\n");
	exit(1);
    }

#ifdef VERBOSE
    fprintf(stdout, "Contact list: %d contacts, ", cont->ncontact);
#endif
}


static void read_reference_contacts(FILE *contactfile, Contl *cont,  Ali *ali)
{
    /* third line contains list of contact pairs */

    int i, position;

    cont->contact = safe_malloc(cont->ncontact * sizeof(int [2]));
    cont->ncontres = safe_malloc(ali->length * sizeof(int));

    for (i = 0; i < ali->length; ++ i)
		cont->ncontres[i] = 0;

    for (i = 0; fscanf(contactfile, "%d", &position) == 1; i ++)
	{
		if ((unsigned) position >= ali->sequence[0].length)
		{
			fprintf(stderr, "Exiting: Contact number out of range\n");
			exit(1);
		}

		if (i / 2 < cont->ncontact)
		{
			cont->contact[i / 2][i % 2] = position;
			++ cont->ncontres[position];
		}
    }

    if (i != 2 * cont->ncontact)
	{
		fprintf(stderr, "Exiting: Expected %d contact numbers; %d read\n", 2 * cont->ncontact, i);
		exit(1);
    }

    /* find average and maximal number of contacts in sequence */
    cont->maxcn = 0;

    for (i = 0; i < ali->length; i ++)
	if (cont->maxcn < cont->ncontres[i])
	    cont->maxcn = cont->ncontres[i];

    cont->avgcn = (float) (2 * cont->ncontact) / ali->length;
#ifdef VERBOSE
    fprintf(stdout, "avg. CN = %3.1f, max. CN = %2d\n", cont->avgcn, cont->maxcn);
#endif
}

static void read_query_contacts(FILE *contactfile, Contl *cont,  Ali *ali)
{
    /* third line contains list of contact pairs */

    int i, position;

    cont->contact = safe_malloc(cont->ncontact * sizeof(int [2]));
    cont->ncontres = safe_malloc(ali->length * sizeof(int));

    for (i = 0; i < ali->length; ++ i)
		cont->ncontres[i] = 0;

    for (i = 0; fscanf(contactfile, "%d", &position) == 1; i ++)
	{
		if ((unsigned) position >= ali->sequence[1].length)
		{
			fprintf(stderr, "Exiting: Contact number out of range\n");
			exit(1);
		}

		if (i / 2 < cont->ncontact)
		{
			cont->contact[i / 2][i % 2] = position;
			++ cont->ncontres[position];
		}
    }

    if (i != 2 * cont->ncontact)
	{
		fprintf(stderr, "Exiting: Expected %d contact numbers; %d read\n", 2 * cont->ncontact, i);
		exit(1);
    }

    /* find average and maximal number of contacts in sequence */
    cont->maxcn = 0;

    for (i = 0; i < ali->length; i ++)
	if (cont->maxcn < cont->ncontres[i])
	    cont->maxcn = cont->ncontres[i];

    cont->avgcn = (float) (2 * cont->ncontact) / ali->length;
#ifdef VERBOSE
    fprintf(stdout, "avg. CN = %3.1f, max. CN = %2d\n", cont->avgcn, cont->maxcn);
#endif
}


void contactlist(FILE *contactfile, Contl *cont,  Ali *ali, int reforque)
{
    if (!reforque)
	{
        read_reference_sequence(contactfile, ali);
		read_nr_contacts(contactfile, cont);
		read_reference_contacts(contactfile, cont, ali);
    }
    else 
	{
        read_query_sequence(contactfile, ali);
		read_nr_contacts(contactfile, cont);
		read_query_contacts(contactfile, cont, ali);
    }
}
