/*==============================================================================
 $Id: structure.c,v 1.3 2005/04/20 14:48:00 jkleinj Exp $ 
 structure.c : Routines for structure operations
 Copyright (C) 2004 Jens Kleinjung
 see README file for Syntax, Copyright and License
==============================================================================*/

#include "aliss.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>
#include "structure.h"

/*----------------------------------------------------------------------------*/
/* read PDB file */

/*
Definition of PDB format:
http://pdbbeta.rcsb.org/pdb/file_formats/pdb/pdbguide2.2/guide2.2_frame.html
The ATOM record:

COLUMNS        DATA TYPE       FIELD         DEFINITION
---------------------------------------------------------------------------------
 1 -  6        Record name     "ATOM  "
 7 - 11        Integer         serial        Atom serial number.
13 - 16        Atom            name          Atom name.
17             Character       altLoc        Alternate location indicator.
18 - 20        Residue name    resName       Residue name.
22             Character       chainID       Chain identifier.
23 - 26        Integer         resSeq        Residue sequence number.
27             AChar           iCode         Code for insertion of residues.
31 - 38        Real(8.3)       x             Orthogonal coordinates for X
                                               in Angstroms.
39 - 46        Real(8.3)       y             Orthogonal coordinates for Y
                                               in Angstroms.
47 - 54        Real(8.3)       z             Orthogonal coordinates for Z
                                               in Angstroms.
55 - 60        Real(6.2)       occupancy     Occupancy.
61 - 66        Real(6.2)       tempFactor    Temperature factor.
73 - 76        LString(4)      segID         Segment identifier,
                                               left-justified.
77 - 78        LString(2)      element       Element symbol,
                                               right-justified.
79 - 80        LString(2)      charge        Charge on the atom.
*/

void read_pdb(FILE *pdbfile, Str *str)
{
	int i, j;
	int k = 0;
	char line[80];
	int allocated = 64;

	/* initialise/allocate memory for set of (64) selected (CA) atom entries */
	str->natom = 0;
	str->atom = safe_malloc(allocated * sizeof(Atom));

	/* allocate memory for sequence residues */
	str->sequence.res = safe_malloc(allocated * sizeof(char));

	/* not all PDB data types are used in this program to save resources */
    while(! feof(pdbfile))
    {
        fgets(line, 80, pdbfile); /* read line */

		/* record name */
		if(strncmp(line, "ATOM  ", 6) != 0)
			continue;

		/* atom number */
		str->atom[str->natom].atom_nr = atoi(&line[6]);

		/* atom name */
		for (i = 12, j = 0; i < 16; )
			str->atom[str->natom].atom_ne[j++] = line[i++];
		str->atom[str->natom].atom_ne[j] = '\0';

		/* decision about recording selected (CA) atoms */
		if(strncmp(str->atom[str->natom].atom_ne, " CA ", 4) != 0)
			continue;

		/* alternative location */
		/*str->atom[str->natom].alt_loc[0] = line[16];
		str->atom[str->natom].alt_loc[1] = '\0';*/

		/* residue name */
		for (i = 17, j = 0; i < 20; )
			str->atom[str->natom].res_ne[j++] = line[i++];
		str->atom[str->natom].res_ne[j] = '\0';

		/* chain identifier */
		str->atom[str->natom].chain_id[0] = line[21];
		str->atom[str->natom].chain_id[1] = '\0';

		/* residue number */
		/*str->atom[str->natom].res_nr = atoi(&line[22]);*/

		/* code for insertion of residues */
		/*str->atom[str->natom].icode[0] = line[26];
		str->atom[str->natom].icode[1] = '\0';*/

		/* coordinates */
		str->atom[str->natom].x = atof(&line[30]);
		str->atom[str->natom].y = atof(&line[38]);
		str->atom[str->natom].z = atof(&line[46]);

		/*printf("x %6.4f, y %6.4f, z %6.4f\n", str->atom[str->natom].x,
			str->atom[str->natom].y, str->atom[str->natom].z);*/

		/* occupancy */
		/*str->atom[str->natom].occupancy = atof(&line[54]);*/

		/* temperature factor */
		/*str->atom[str->natom].temp_f = atof(&line[60]);*/

		/* segment identifier */
		/*for (i = 72, j = 0; i < 76; )
			str->atom[str->natom].seg_id[j++] = line[i++];
		str->atom[str->natom].seg_id[j] = '\0';*/

		/* element */
		/*for (i = 76, j = 0; i < 78; )
			str->atom[str->natom].element[j++] = line[i++];
		str->atom[str->natom].element[j] = '\0';*/

		/* charge */
		/*for (i = 78, j = 0; i < 80; )
			str->atom[str->natom].charge[j++] = line[i++];
		str->atom[str->natom].charge[j] = '\0';*/

		/* description: everything before coordinates */
		for (i = 0, j = 0; i < 30; )
			str->atom[str->natom].descrip[j++] = line[i++];
		str->atom[str->natom].descrip[j] = '\0';

		/* assign residue to sequence and count selected (CA) atoms */
		/* transform to 1-letter code and store sequence */
		str->sequence.res[k++] = aacode(str->atom[str->natom].res_ne);
		++ str->natom;

		/* allocate more memory if needed */
		if (str->natom == allocated)
		{
			allocated += 64;
			str->atom = safe_realloc(str->atom, allocated * sizeof(Atom));
			str->sequence.res = safe_realloc(str->sequence.res, allocated * sizeof(char));
		}
	}
	str->sequence.res[k] = '\0';

#ifdef VERBOSE
	fprintf(stdout, "Structure '%s', length = %5d\n", str->sequence.name, str->natom);
#endif
}

/*----------------------------------------------------------------------------*/
/* read all atoms from PDB file */
void read_all_pdb(FILE *pdbfile, Str *str)
{
	int i, j;
	char line[80];
	int allocated = 64;

	/* initialise/allocate memory for set of (64) all atom entries */
	str->nallatom = 0;
	str->allatom = safe_malloc(allocated * sizeof(Atom));

	/* reset file pointer */
	fseek(pdbfile, 0, 0);

	/* for comments see routine above */
    while(! feof(pdbfile))
    {
        fgets(line, 80, pdbfile); /* read line */

		if(strncmp(line, "ATOM  ", 6) != 0)
			continue;

		str->allatom[str->nallatom].atom_nr = atoi(&line[6]);

		for (i = 12, j = 0; i < 16; )
			str->allatom[str->nallatom].atom_ne[j++] = line[i++];
		str->allatom[str->nallatom].atom_ne[j] = '\0';

		for (i = 17, j = 0; i < 20; )
			str->allatom[str->nallatom].res_ne[j++] = line[i++];
		str->allatom[str->nallatom].res_ne[j] = '\0';

		str->allatom[str->nallatom].chain_id[0] = line[21];
		str->allatom[str->nallatom].chain_id[1] = '\0';

		str->allatom[str->nallatom].x = atof(&line[30]);
		str->allatom[str->nallatom].y = atof(&line[38]);
		str->allatom[str->nallatom].z = atof(&line[46]);

		for (i = 0, j = 0; i < 30; )
			str->allatom[str->nallatom].descrip[j++] = line[i++];
		str->allatom[str->nallatom].descrip[j] = '\0';

		++ str->nallatom;

		if (str->nallatom == allocated)
		{
			allocated += 64;
			str->allatom = safe_realloc(str->allatom, allocated * sizeof(Atom));
		}
	}

#ifdef VERBOSE
	fprintf(stdout, "\tall atom number = %5d\n", str->nallatom);
#endif
}

/*----------------------------------------------------------------------------*/
/* amino acid 3-letter to 1-letter code conversion */
char aacode(char *code3)
{
	int i;
	char aa1 = 'O';
	char *aa3[] = {"ALA","---","CYS","ASP","GLU","PHE","GLY","HIS","ILE","---","LYS","LEU","MET","ASN","---","PRO","GLN","ARG","SER","THR","---","VAL","TRP","---","TYR","---"};

	for (i = 0; i < 26; ++i)
		if (strncmp(code3, aa3[i], 3) == 0)
		{
			aa1 = i + 65;
			break;
		}

	if (aa1 == 'O')
	{
		fprintf(stderr, "Exiting: Problematic residue type '%3s' in protein structure.\n"
						"Solution: Change to standard type or implement 1-letter code conversion.\n",
						code3);
		exit(1);
	}
	
	return aa1;
}

/*----------------------------------------------------------------------------*/
/* write superimposed PDB files */
/* The CONECT record */
/*
COLUMNS         DATA TYPE        FIELD           DEFINITION
---------------------------------------------------------------------------------
 1 -  6         Record name      "CONECT" 
 7 - 11         Integer          serial          Atom serial number
12 - 16         Integer          serial          Serial number of bonded atom
17 - 21         Integer          serial          Serial number of bonded atom
22 - 26         Integer          serial          Serial number of bonded atom
27 - 31         Integer          serial          Serial number of bonded atom
32 - 36         Integer          serial          Serial number of hydrogen bonded atom 
37 - 41         Integer          serial          Serial number of hydrogen bonded atom 
42 - 46         Integer          serial          Serial number of salt bridged atom
47 - 51         Integer          serial          Serial number of hydrogen bonded atom
52 - 56         Integer          serial          Serial number of hydrogen bonded atom
57 - 61         Integer          serial          Serial number of salt bridged atom
*/

void write_pdb(FILE *pdboutfile, Str *pdb0, Str *pdb1, Ali *ali, int allatom)
{
	int i;
	/*int z = 0;*/
	int natom0 = 0;
	int natom1 = 0;
	Atom *patom0 = 0;
	Atom *patom1 = 0;

	if (allatom)
	{
		natom0 = pdb0->nallatom;
		patom0 = pdb0->allatom;
		natom1 = pdb1->nallatom;
		patom1 = pdb1->allatom;
	}
	else
	{
		natom0 = pdb0->natom;
		patom0 = pdb0->atom;
		natom1 = pdb1->natom;
		patom1 = pdb1->atom;
	}

	/* pdb0 */
	for (i = 0; i < natom0; ++i)
	{
		fprintf(pdboutfile, "%30s", patom0[i].descrip);
		fprintf(pdboutfile, "%8.3f%8.3f%8.3f\n", 
			patom0[i].x, patom0[i].y, patom0[i].z);
	}
	fprintf(pdboutfile, "TER\n");

	/* pdb1 : transformed */
	for (i = 0; i < natom1; ++i)
	{
		fprintf(pdboutfile, "%30s", patom1[i].descrip);
		fprintf(pdboutfile, "%8.3f%8.3f%8.3f\n", 
			patom1[i].tx, patom1[i].ty, patom1[i].tz);
	}
	fprintf(pdboutfile, "TER\n");

	fprintf(pdboutfile, "END\n");
}

/*----------------------------------------------------------------------------*/
/* define fragment points to superimpose */
void def_points(int first0, int first1, unsigned int pep, gsl_matrix *X, gsl_matrix *Y, Str *pdb0, Str *pdb1)
{
	int i;

	for (i = 0; i < pep; ++i)
	{
		/* matrix Y contains points to move to = template */
		gsl_matrix_set(Y, i, 0, pdb0->atom[i+first0].x);
		gsl_matrix_set(Y, i, 1, pdb0->atom[i+first0].y);
		gsl_matrix_set(Y, i, 2, pdb0->atom[i+first0].z);

		/* matrix X contains points to be moved = query */
		gsl_matrix_set(X, i, 0, pdb1->atom[i+first1].x);
		gsl_matrix_set(X, i, 1, pdb1->atom[i+first1].y);
		gsl_matrix_set(X, i, 2, pdb1->atom[i+first1].z);
	}
}

/*----------------------------------------------------------------------------*/
/* define points from alignment to superimpose */
void def_ali_points(Ali *ali, unsigned int pep, gsl_matrix *X, gsl_matrix *Y, Str *pdb0, Str *pdb1)
{
	int i;

	for (i = 0; i < pep; ++i)
	{
		/* matrix Y contains points to move to = template */
		gsl_matrix_set(Y, i, 0, pdb0->atom[ali->sequence[0].pos[i+1]].x);
		gsl_matrix_set(Y, i, 1, pdb0->atom[ali->sequence[0].pos[i+1]].y);
		gsl_matrix_set(Y, i, 2, pdb0->atom[ali->sequence[0].pos[i+1]].z);

		/* matrix X contains points to be moved = query */
		gsl_matrix_set(X, i, 0, pdb1->atom[ali->sequence[1].pos[i+1]].x);
		gsl_matrix_set(X, i, 1, pdb1->atom[ali->sequence[1].pos[i+1]].y);
		gsl_matrix_set(X, i, 2, pdb1->atom[ali->sequence[1].pos[i+1]].z);
	}
}

/*----------------------------------------------------------------------------*/
/* calculate logg-odds structure score */
float struc_score(int pep, float rmsd, float stscf)
{
	float stsc; /* structure score */

	if (rmsd < 0.0001) /* minimum value for rmsd, otherwise stsc towards inf */
		rmsd = 0.0001;
	stsc = log(stscf * pep / rmsd);	
	return stsc;
}

/*----------------------------------------------------------------------------*/
/* transform coordinates of query structure pdb1 : x' = Ux + t */
/*   where 'U' is the rotation matrix and 't' is the translation vector */
/*   the values of which were determined in the Kabsch routine */
float transform_coordinates(int first0, int first1, int natom, gsl_matrix *U, gsl_vector *t, Str *pdb0, Str *pdb1)
{
	int i;
	float rmsd = 0.;

	for (i = 0; i < natom; ++i) /* for natom atoms */
	{
		/* transform coordinates */
		pdb1->atom[i+first1].tx =	pdb1->atom[i+first1].x * gsl_matrix_get(U,0,0) +
									pdb1->atom[i+first1].y * gsl_matrix_get(U,0,1) +
									pdb1->atom[i+first1].z * gsl_matrix_get(U,0,2) +
									gsl_vector_get(t,0); 
		pdb1->atom[i+first1].ty =	pdb1->atom[i+first1].x * gsl_matrix_get(U,1,0) +
									pdb1->atom[i+first1].y * gsl_matrix_get(U,1,1) +
									pdb1->atom[i+first1].z * gsl_matrix_get(U,1,2) +
									gsl_vector_get(t,1); 
		pdb1->atom[i+first1].tz =	pdb1->atom[i+first1].x * gsl_matrix_get(U,2,0) +
									pdb1->atom[i+first1].y * gsl_matrix_get(U,2,1) +
									pdb1->atom[i+first1].z * gsl_matrix_get(U,2,2) +
									gsl_vector_get(t,2); 

		/* calculate rmsd */
		rmsd += sqrt(pow(pdb0->atom[i+first0].x - pdb1->atom[i+first1].tx, 2) 
				   + pow(pdb0->atom[i+first0].y - pdb1->atom[i+first1].ty, 2)
				   + pow(pdb0->atom[i+first0].z - pdb1->atom[i+first1].tz, 2));
	}
	rmsd /= pdb0->natom;

	return rmsd;
}

/*----------------------------------------------------------------------------*/
/* the same as above, but transformation of all atoms of pdb1 */
/*   and with an RMSD calculation for selected atoms = aligned residues */
float transform_ali_coordinates(Ali*ali, gsl_matrix *U, gsl_vector *t, Str *pdb0, Str *pdb1, int allatom)
{
	int i = 0;
	float rmsd = 0.;
	int pos0 = 0;
	int pos1 = 0;

	/* superpositioning of Calpha atoms */
	for (i = 0; i < ali->sequence[1].length; ++i) /* for all query atoms */
	{
		/* transform all coordinates of query sequence */
		/* 'first' refers to residue number of N-terminus in sequence alignment */
		pdb1->atom[i].tx =	pdb1->atom[i].x * gsl_matrix_get(U,0,0) +
							pdb1->atom[i].y * gsl_matrix_get(U,0,1) +
							pdb1->atom[i].z * gsl_matrix_get(U,0,2) +
							gsl_vector_get(t,0); 
		pdb1->atom[i].ty =	pdb1->atom[i].x * gsl_matrix_get(U,1,0) +
							pdb1->atom[i].y * gsl_matrix_get(U,1,1) +
							pdb1->atom[i].z * gsl_matrix_get(U,1,2) +
							gsl_vector_get(t,1); 
		pdb1->atom[i].tz =	pdb1->atom[i].x * gsl_matrix_get(U,2,0) +
							pdb1->atom[i].y * gsl_matrix_get(U,2,1) +
							pdb1->atom[i].z * gsl_matrix_get(U,2,2) +
							gsl_vector_get(t,2); 
	}

	/* superpositioning of all atoms if selected */
	if (allatom)
	{
		for (i = 0; i < pdb1->nallatom; ++i) /* for all query atoms */
		{
			/* transform all coordinates of query sequence */
			/* 'first' refers to residue number of N-terminus in sequence alignment */
			pdb1->allatom[i].tx =	pdb1->allatom[i].x * gsl_matrix_get(U,0,0) +
									pdb1->allatom[i].y * gsl_matrix_get(U,0,1) +
									pdb1->allatom[i].z * gsl_matrix_get(U,0,2) +
									gsl_vector_get(t,0); 
			pdb1->allatom[i].ty =	pdb1->allatom[i].x * gsl_matrix_get(U,1,0) +
									pdb1->allatom[i].y * gsl_matrix_get(U,1,1) +
									pdb1->allatom[i].z * gsl_matrix_get(U,1,2) +
									gsl_vector_get(t,1); 
			pdb1->allatom[i].tz =	pdb1->allatom[i].x * gsl_matrix_get(U,2,0) +
									pdb1->allatom[i].y * gsl_matrix_get(U,2,1) +
									pdb1->allatom[i].z * gsl_matrix_get(U,2,2) +
									gsl_vector_get(t,2); 
		}
	}

	/* rmsd calculation of the above superpositioning based on selected atoms */
	for (i = 1; i < ali->sequence[1].pos[0]; ++i) /* for all selected atoms */
	{
		pos0 = ali->sequence[0].pos[i];
		pos1 = ali->sequence[1].pos[i];

		/* calculate rmsd */
		rmsd +=  sqrt(pow(pdb0->atom[pos0].x - pdb1->atom[pos1].tx, 2) 
				   + pow(pdb0->atom[pos0].y - pdb1->atom[pos1].ty, 2)
				   + pow(pdb0->atom[pos0].z - pdb1->atom[pos1].tz, 2));
	}
	rmsd /= ali->sequence[1].pos[0];

	return rmsd;
}

/*----------------------------------------------------------------------------*/
/* assign sequences from structure record to sequence record */
/* the DP routine operates on the sequence record */
void assign_seq(Ali *ali, Str *pdb0, Str *pdb1)
{
	/* assign residues */
    ali->sequence[0].res = safe_malloc((pdb0->natom + 1) * sizeof(char));
    strcpy(ali->sequence[0].res, pdb0->sequence.res);
    ali->sequence[0].length = strlen(pdb0->sequence.res);

    ali->sequence[1].res = safe_malloc((pdb1->natom + 1) * sizeof(char));
    strcpy(ali->sequence[1].res, pdb1->sequence.res);
    ali->sequence[1].length = strlen(pdb1->sequence.res);
}
